// qStat.cpp: dfinit le point d'entre pour l'application console.
//

#include "stdafx.h"
#ifdef linux
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include <errno.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#endif


// to reuse the structures for the total
#define SUMINDEX 1000

void exitHelp();

double pi = atan(1.0)*4.0;

int aveR(double f1)
{
	return (int)(floor(f1 + 0.5));

}

int fgetsDI(FILE *fcb, double & d1, int & n2)
{
	char chaine[512], *pc = chaine;
	if (fgets(chaine, 512, fcb))
	{
		char* pEnd;
		d1 = floor(strtod(chaine, &pEnd));
		n2 = (int)floor(strtod(pEnd, NULL));
		return 1;
	}
	return 0;
}


class dataAngles
{
public:
	int Angle;
	int lines;
	int tot_plus;
	int tot_minus;
	int tot_undetected;

	double detection_rate;
	double part_plus;
	double part_minus;

	dataAngles(int angle)
	{
		Angle = angle;
		tot_plus = tot_minus = tot_undetected = 0;
	};

	void Add(int pol)
	{
		lines++;
		if (pol > 0)
			tot_plus++;
		else if (pol < 0)
			tot_minus++;
		else
			tot_undetected++;
	};

	void finalize()
	{
		if (lines && (tot_plus || tot_minus))
		{
			double tot = (double)(tot_plus + tot_minus);
			detection_rate = 100.0*tot / (double)lines;
			part_plus = 100.0*(double)tot_plus / tot;
			part_minus = 100.0*(double)tot_minus / tot;
		}
		else
		{
			detection_rate =
				part_plus =
				part_minus = 0.0;
		}
	};


	void TotaliseTo(class dataAngles * total)
	{
		total->lines += lines;
		total->tot_plus += tot_plus;
		total->tot_minus += tot_minus;
		total->tot_undetected += tot_undetected;
	};


	void Header(FILE *fcb)
	{
		fprintf(fcb, "Angle\tUndetected\t%%\tDetected\t%%\tDetected +\t%%\tDetected -\t%%\n");
	};


	void outLine(FILE *fcb)
	{
		if (Angle == SUMINDEX)
		{
			fprintf(fcb, "Sums\t");
		}
		else
		{
			fprintf(fcb, "%d\t", Angle);
		}

		fprintf(fcb, "%d\t%.02f\t%d\t%.02f\t%d\t%.02f\t%d\t%.02f\n",
			tot_undetected, 100.0 - detection_rate,
			tot_plus + tot_minus, detection_rate,
			tot_plus, part_plus,
			tot_minus, part_minus);
	};
};

class dataAngles * dA[361];
class dataAngles * AllAng;



class dataDiffAngles
{
public:
	int Diff;
	double q_correlation;
	int lines;
	int p_pairs;
	int p_pairs_eq;
	int p_pairs_ne;
	int p_singles;
	int p_none;



	double s_correlation;
	double sq_diff;
	double e_csch;


	double detection_p_rate;
	double detection_s_rate;
	double detection_n_rate;

	dataDiffAngles(int diff)
	{
		Diff = diff;
		lines = p_pairs = p_pairs_eq = p_pairs_ne = p_singles = p_none = 0;
		sq_diff = 0.0;
	};
	~dataDiffAngles() {};

	void Add(int pol_1, int pol_2)
	{
		lines++;
		if (pol_1 == 0 && pol_2 == 0)
		{
			p_none++;
		}
		else if (pol_1 == 0 || pol_2 == 0)
		{
			p_singles++;
		}
		else if (pol_1 == pol_2)
		{
			p_pairs_eq++;
			p_pairs++;
		}
		else
		{
			p_pairs_ne++;
			p_pairs++;
		}
	};

	void finalize()
	{
		if (Diff != SUMINDEX)
		{
			q_correlation = cos(Diff* pi / 180.0);
			q_correlation *= q_correlation;
			if (p_pairs)
				e_csch = (double)(p_pairs_eq - p_pairs_ne) / (double)p_pairs;
			else
				e_csch = 0;
		}
		else
		{
			q_correlation = 0.5;
			e_csch = 0;
		}
		if (lines)
		{
			detection_p_rate = 100.0*(double)p_pairs / (double)lines;
			detection_s_rate = 100.0*(double)p_singles / (double)lines;
			detection_n_rate = 100.0*(double)p_none / (double)lines;
			if (p_pairs)
			{
				s_correlation = (double)p_pairs_eq / (double)p_pairs;
			}
			else
			{
				s_correlation = 0;
			}
		}
		else
		{
			s_correlation =
				detection_p_rate =
				detection_s_rate =
				detection_n_rate = 0.0;
		}


		if (Diff != SUMINDEX)
		{
			sq_diff = fabs(s_correlation - q_correlation);
		}
		else
		{
			// leave the total untouched ;
		}
	};

	void TotaliseTo(class dataDiffAngles * total)
	{
		total->lines += lines;
		total->p_pairs += p_pairs;
		total->p_pairs_eq += p_pairs_eq;
		total->p_pairs_ne += p_pairs_ne;
		total->p_singles += p_singles;
		total->p_none += p_none;
		total->sq_diff += sq_diff;
	};

	void Header(FILE *fcb)
	{
		fprintf(fcb, "Delta Rotat.\tEmitted\tPairs\tequal\tnot equal\tSingles only\tNot detected\tproba MQ\tSimulation\tDiff\tE CSCH\tPairs\tSingles\tNot detected\n");
	};

	void outLine(FILE *fcb)
	{
		if (Diff == SUMINDEX)
		{
			fprintf(fcb, "Sums\t");
		}
		else
		{
			fprintf(fcb, "%d\t", Diff);
		}

		fprintf(fcb, "%d\t%d\t%d\t%d\t%d\t%d\t%.05f\t%.05f\t%.05f\t%.05f\t%.02f\t%.02f\t%.02f\n",
			// tot detections and correlations
			lines, p_pairs, p_pairs_eq, p_pairs_ne, p_singles, p_none,
			// ratios and ref
			q_correlation, s_correlation, sq_diff, e_csch, detection_p_rate, detection_s_rate, detection_n_rate);
	};
};

class dataDiffAngles * dDA[361];
class dataDiffAngles * AllDiff;

void exitHelp()
{
	printf("outputs the Alice and Bob after-polarizer files comparison for the numerical EPR experiment :\n");
	printf("# qStat alice_polarized_file bob_polarized_file output_file\n");
	printf("Both arguments are mandatory\n");
	printf("0 < number_of_sections < 50\n");
	exit(1);
}



int main(int argc, char *argv[])
{
	char	fichierINALICE[512],
		fichierINBOB[512],
		fichierOUT[512];

	FILE  *fcbINALICE, *fcbINBOB, *fcbOUT;

	// extract arguments

	if (argc != 4 )
	{
		exitHelp();
	}

	strcpy(fichierINALICE, argv[1]);
	if (!*fichierINALICE)
	{
		exitHelp();
	}

	strcpy(fichierINBOB, argv[2]);
	if (!*fichierINBOB)
	{
		exitHelp();
	}

	strcpy(fichierOUT, argv[3]);
	if (!*fichierOUT)
	{
		exitHelp();
	}


	// ################################################################################################

	// opening files
	//  *fcbINALICE, *fcbINBOB, *fcbOUT;

	if (!(fcbOUT = fopen(fichierOUT, "w")))
	{
		printf("unable to create the result file %s\n", fichierOUT);
		exit(1);
	}

	if (!(fcbINALICE = fopen(fichierINALICE, "r")))
	{
		printf("unable to read the ALICE file %s\n", fichierINALICE);
		fclose(fcbOUT);
		exit(1);
	}

	if (!(fcbINBOB = fopen(fichierINBOB, "r")))
	{
		printf("unable to read the BOB file %s\n", fichierINBOB);
		fclose(fcbINALICE);
		fclose(fcbOUT);
		exit(1);
	}

	int i;
	for (i = 0; i < 361; i++)
		dDA[i] = new class dataDiffAngles(i - 180);
	AllDiff = new class dataDiffAngles(SUMINDEX);

	for (i = 0; i < 181; i++)
		dA[i] = new class dataAngles(i - 90);
	AllAng = new class dataAngles(SUMINDEX);

	// seed the randomizer
	srand((unsigned int)time(NULL)); //  seed48()  and  lcong48()

	// now arguments are ready and files are open

	int nblignes = 0, pol_1, pol_2;
	double frotation_1, frotation_2;
	while (fgetsDI(fcbINALICE, frotation_1, pol_1) && fgetsDI(fcbINBOB, frotation_2, pol_2))
	{
		nblignes++;
		// got the pairs of rotations and results , let's do some minimalist stats

		// total by angle diff
		dDA[aveR(frotation_1 - frotation_2) + 180]->Add(pol_1, pol_2);

		// total by angle for each arm
		dA[aveR(frotation_1) + 90]->Add(pol_1);
		dA[aveR(frotation_2) + 90]->Add(pol_2);

	}
	fclose(fcbINALICE);
	fclose(fcbINBOB);

	for (i = 1; i < 360; i++)
	{
		dDA[i]->finalize();
		dDA[i]->TotaliseTo(AllDiff);
	}

	AllDiff->finalize();

	for (i = 0; i < 180; i++)
	{
		dA[i]->finalize();
		dA[i]->TotaliseTo(AllAng);
	}
	AllAng->finalize();

	//---- output --------------------------------------
	AllDiff->Header(fcbOUT);
	AllDiff->outLine(fcbOUT);
	for (i = 1; i < 360; i++)
	{
		dDA[i]->outLine(fcbOUT);
	}
	fprintf(fcbOUT, "\n\n\n");

	AllAng->Header(fcbOUT);
	AllAng->outLine(fcbOUT);

	for (i = 0; i < 180; i++)
	{
		dA[i]->outLine(fcbOUT);
	}

	fprintf(fcbOUT, "\n\n\n");

	fprintf(fcbOUT, "\tBell and CSCH for 22,22,22,66\n");
	fprintf(fcbOUT, "Bell Simu:\t3 * E(22) - E(66)\t\t\t\t\t%.05f\n",
		3 * dDA[22 + 180]->s_correlation - dDA[66 + 180]->s_correlation);
	fprintf(fcbOUT, "CSCH Simu:\t3 * E_csch(22) - E_csch(66)\t\t\t\t\t%.05f\n",
		3 * dDA[22 + 180]->e_csch - dDA[66 + 180]->e_csch);

	fprintf(fcbOUT, "\n\n\n");

	fprintf(fcbOUT, "\tBell and CSCH for 23,23,23,69\n");
	fprintf(fcbOUT, "Bell Simu:\t3 * E(23) - E(69)\t\t\t\t\t%.05f\n",
		3 * dDA[23 + 180]->s_correlation - dDA[69 + 180]->s_correlation);
	fprintf(fcbOUT, "CSCH Simu:\t3 * E_csch(23) - E_csch(69)\t\t\t\t\t%.05f\n",
		3 * dDA[23 + 180]->e_csch - dDA[69 + 180]->e_csch);

	fprintf(fcbOUT, "\n\n\n");
	fprintf(fcbOUT, "\tBell and CSCH for 24,24,24,72\n");
	fprintf(fcbOUT, "Bell Simu:\t3 * E(24) - E(72)\t\t\t\t\t%.05f\n",
		3 * dDA[24 + 180]->s_correlation - dDA[72 + 180]->s_correlation);
	fprintf(fcbOUT, "CSCH Simu:\t3 * E_csch(24) - E_csch(72)\t\t\t\t\t%.05f\n",
		3 * dDA[24 + 180]->e_csch - dDA[72 + 180]->e_csch);

	fprintf(fcbOUT, "\n\n\n");

	fclose(fcbOUT);
	exit(0);

}
